{ File   : CPS2BMP.dpr
  Author : Deniz Oezmen
  Created: 2007-03-19
  Changed: 2007-03-20

  Kyrandia trilogy CPS converter, based on ScummVM sources. Thus:


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
program CPS2BMP;

{$APPTYPE CONSOLE}

uses
  SysUtils, SaveBMP3;

type
  TCPSHeader = packed record
                 DataSize  : Word;
                 CompType,
                 Unknown   : Byte;
                 UncompSize: Longword;
                 PalSize   : Word;
               end;

  TBuffer    = array of Byte;

var
  ExpandPal: Boolean = False;

function OpenCPSFile(var CPSFile: file; CPSName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(CPSFile, CPSName);

  {$I-}
  Reset(CPSFile, 1);
  {$I+}
  OpenCPSFile := (IOResult = 0)
end;

function ReadHeader(var CPSFile: file; out CPSHeader: TCPSHeader): Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(CPSFile, CPSHeader, SizeOf(TCPSHeader), BytesRead);
  ReadHeader := (SizeOf(TCPSHeader) = BytesRead) and
    (CPSHeader.CompType in [0, 3, 4])
end;

procedure SetupDefaultPalette(out Palette: TPalette3);
var
  i: Byte;
begin
  for i := 0 to 255 do
    with Palette[i] do
    begin
      B := i;
      G := i;
      R := i
    end
end;

procedure ReadCPSPalette(var CPSFile: file; const CPSHeader: TCPSHeader;
  var Palette: TPalette3);
var
  i   : Longint;
  s, t: Byte;
begin
  if CPSHeader.PalSize = 0 then
    Exit;

  WriteLn('  found internal palette');

  BlockRead(CPSFile, Palette[0], CPSHeader.PalSize);

  if ExpandPal then
    s := 2
  else
    s := 0;

  for i := 0 to CPSHeader.PalSize div 3 - 1 do
    with Palette[i] do
    begin
      t := B shl s;
      B := R shl s;
      G := G shl s;
      R := t
    end
end;

procedure DecodeType0(const InBuffer: TBuffer; var OutBuffer: TBuffer);
begin
  if Length(OutBuffer) < Length(InBuffer) then
    Exit;

  Move(InBuffer[0], OutBuffer[0], Length(InBuffer))
end;

procedure DecodeType3(const InBuffer: TBuffer; var OutBuffer: TBuffer);
var
  InIdx,
  OutIdx: Longint;
  OpCode: Shortint;
  Size  : Word;
begin
  InIdx := Low(InBuffer);
  OutIdx := Low(OutBuffer);

  while InIdx <= High(InBuffer) do
  begin
    OpCode := InBuffer[InIdx];
    Inc(InIdx);

    if OpCode = 0 then
    begin
      if InIdx + 2 > High(InBuffer) then
        Exit;

      Size := InBuffer[InIdx] shl 8 + InBuffer[InIdx + 1];
      Inc(InIdx, 2);

      if OutIdx + Size - 1 > High(OutBuffer) then
        Exit;

      FillChar(OutBuffer[OutIdx], Size, InBuffer[InIdx]);
      Inc(InIdx);
      Inc(OutIdx, Size)
    end
    else if OpCode < 0 then
    begin
      OpCode := -OpCode;

      if (InIdx > High(InBuffer)) or
        (OutIdx + OpCode - 1 > High(OutBuffer)) then
        Exit;

      FillChar(OutBuffer[OutIdx], OpCode, InBuffer[InIdx]);
      Inc(InIdx, 1);
      Inc(OutIdx, OpCode)
    end
    else
    begin
      if (InIdx + OpCode - 1 > High(InBuffer)) or
        (OutIdx + OpCode - 1 > High(OutBuffer)) then
        Exit;

      Move(InBuffer[InIdx], OutBuffer[OutIdx], OpCode);
      Inc(InIdx, OpCode);
      Inc(OutIdx, OpCode)
    end
  end
end;

procedure DecodeType4(const InBuffer: TBuffer; var OutBuffer: TBuffer);
var
  i,
  InIdx,
  OutIdx: Longint;
  OpCode: Byte;
  Size,
  Offset: Word;
begin
  InIdx := Low(InBuffer);
  OutIdx := Low(OutBuffer);

  while InIdx <= High(InBuffer) do
  begin
    OpCode := InBuffer[InIdx];
    Inc(InIdx);

    if (OpCode and $80) = 0 then
    begin
      if InIdx > High(InBuffer) then
        Exit;

      Size := (OpCode shr 4) + 3;
      Offset := ((OpCode and $0f) shl 8) or InBuffer[InIdx];
      Inc(InIdx);

      if (OutIdx - Offset < Low(OutBuffer)) or
        (OutIdx + Size - 1 > High(OutBuffer)) then
        Exit;

      for i := 0 to Size - 1 do
        OutBuffer[OutIdx + i] := OutBuffer[OutIdx - Offset + i];
      Inc(OutIdx, Size)
    end
    else if (OpCode and $40) <> 0 then
    begin
      if OpCode = $fe then
      begin
        if InIdx + 1 > High(InBuffer) then
          Exit;

        Move(InBuffer[InIdx], Size, 2);
        Inc(InIdx, 2);

        if (InIdx > High(InBuffer)) or
          (OutIdx + Size - 1 > High(OutBuffer)) then
          Exit;

        FillChar(OutBuffer[OutIdx], Size, InBuffer[InIdx]);
        Inc(InIdx);
        Inc(OutIdx, Size)
      end
      else
      begin
        if OpCode = $ff then
        begin
          if InIdx + 1 > High(InBuffer) then
            Exit;

          Move(InBuffer[InIdx], Size, 2);
          Inc(InIdx, 2)
        end
        else
          Size := (OpCode and $3f) + 3;

        if InIdx + 1 > High(InBuffer) then
          Exit;

        Move(InBuffer[InIdx], Offset, 2);
        Inc(InIdx, 2);

        if (Low(OutBuffer) + Offset + Size > High(OutBuffer)) or
          (OutIdx + Size - 1 > High(OutBuffer)) then
          Exit;

        for i := 0 to Size - 1 do
          OutBuffer[OutIdx + i] := OutBuffer[Low(OutBuffer) + Offset + i];
        Inc(OutIdx, Size)
      end
    end
    else if OpCode <> $80 then
    begin
      Size := OpCode and $3f;

      if (InIdx + Size - 1 > High(InBuffer)) or
        (OutIdx + Size - 1 > High(OutBuffer)) then
        Exit;
      Move(InBuffer[InIdx], OutBuffer[OutIdx], Size);
      Inc(InIdx, Size);
      Inc(OutIdx, Size)
    end
    else
      Exit
  end
end;

procedure ReadPaletteFile(var PalFile: file; var Palette: TPalette3);
var
  i, t, s  : Byte;
  PalHeader: TCPSHeader;
  InBuffer,
  OutBuffer: TBuffer;
begin
  if FileSize(PalFile) = SizeOf(Palette) then
    BlockRead(PalFile, Palette[0], SizeOf(Palette))
  else
  begin
    if not ReadHeader(PalFile, PalHeader) then
      Exit;

    if (PalHeader.UncompSize <> SizeOf(Palette)) or
      (PalHeader.PalSize <> 0) then
      Exit;

    SetLength(InBuffer, FileSize(PalFile) - FilePos(PalFile));
    SetLength(OutBuffer, PalHeader.UncompSize);

    BlockRead(PalFile, InBuffer[0], Length(InBuffer));

    case PalHeader.CompType of
      0: DecodeType0(InBuffer, OutBuffer);
      3: DecodeType3(InBuffer, OutBuffer);
      4: DecodeType4(InBuffer, OutBuffer)
    else
      Exit
    end;

    Move(OutBuffer[0], Palette, SizeOf(Palette))
  end;

  if ExpandPal then
    s := 2
  else
    s := 0;

  for i := 0 to 255 do
    with Palette[i] do
    begin
      t := B shl s;
      B := R shl s;
      G := G shl s;
      R := t
    end
end;

procedure Convert(var CPSFile: file; const CPSName: string;
  const CPSHeader: TCPSHeader; Palette: TPalette3; SearchPal: Boolean);
var
  InBuffer,
  OutBuffer: TBuffer;
  PalFile  : file;
begin
  if CPSHeader.UncompSize <> 64000 then
    Exit;

  if SearchPal then
    if OpenCPSFile(PalFile, ChangeFileExt(CPSName, '.col')) then
    begin
      ReadPaletteFile(PalFile, Palette);
      CloseFile(PalFile);
      WriteLn('  found external palette')
    end;

  ReadCPSPalette(CPSFile, CPSHeader, Palette);

  SetLength(InBuffer, FileSize(CPSFile) - FilePos(CPSFile));
  SetLength(OutBuffer, CPSHeader.UncompSize);

  BlockRead(CPSFile, InBuffer[0], Length(InBuffer));

  case CPSHeader.CompType of
    0: DecodeType0(InBuffer, OutBuffer);
    3: DecodeType3(InBuffer, OutBuffer);
    4: DecodeType4(InBuffer, OutBuffer)
  else
    Exit
  end;

  Dump8BitBMP(CPSName + '.bmp', 320, 200, Palette, @OutBuffer[0]);
end;

procedure ParseFiles(FilePara, PalName: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  CPSName  : string;
  CPSFile,
  PalFile  : file;
  SearchPal: Boolean;
  CPSHeader: TCPSHeader;
  Palette  : TPalette3;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    SetupDefaultPalette(Palette);

    SearchPal := True;
    if PalName <> '' then
    begin
      if OpenCPSFile(PalFile, PalName) then
      begin
        ReadPaletteFile(PalFile, Palette);
        CloseFile(PalFile);
        SearchPal := False
      end
      else
        WriteLn(Format('Warning: Could not open palette file "%s".',
          [PalName]))
    end;

    repeat
      FileName := SearchRec.Name;
      CPSName := Dir + FileName;

      WriteLn(CPSName);

      // if this file is accessible, analyze it
      if OpenCPSFile(CPSFile, CPSName) then
      begin
        if ReadHeader(CPSFile, CPSHeader) then
          Convert(CPSFile, CPSName, CPSHeader, Palette, SearchPal)
        else
          WriteLn('  not a valid (or supported) CPS file');
        CloseFile(CPSFile)
      end
      else
        Writeln(CPSName, '  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, '  not found')
end;

begin
  if ParamCount >= 1 then
    begin
      if AnsiLowerCase(ParamStr(2)) = '/kyra1' then
      begin
        ExpandPal := True;
        ParseFiles(ParamStr(1), '')
      end
      else
      begin
        if AnsiLowerCase(ParamStr(3)) = '/kyra1' then
          ExpandPal := True;
        ParseFiles(ParamStr(1), ParamStr(2))
      end
    end
  else
    Writeln('Incorrect number of parameters.')
end.

