{ File   : PAKExt.dpr
  Author : Deniz Oezmen
  Created: 2007-03-19
  Changed: 2010-09-01

  Kyrandia trilogy PAK extractor.
}
program PAKExt;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  SysUtils;

type
  TPAKEntry = record
                Name  : string;
                Offset,
                Size  : Longint;
              end;

function OpenInFile(var InFile: file; const InName: AnsiString): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; const OutName: AnsiString): Boolean;
begin
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

procedure ExtractFiles(var PAKFile: file);
var
  i, j,
  Idx     : Longint;
  OutFile : file;
  c       : Char;
  Buffer  : array[0..32767] of Byte;
  FileList: array of TPAKEntry;
begin
  SetLength(FileList, 0);

  repeat
    Idx := Length(FileList);
    SetLength(FileList, Idx + 1);
    with FileList[Idx] do
    begin
      {$I-}
      BlockRead(PAKFile, Offset, SizeOf(Offset));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('  error reading structures');
        SetLength(FileList, 0);
        Exit
      end;

      // certain game editions have a zero offset in their PAK files to denote
      // the directory end
      if Offset = 0 then
        Offset := FileSize(PAKFile)
      else
      begin
        Name := '';
        repeat
          {$I-}
          BlockRead(PAKFile, c, 1);
          {$I+}
          if IOResult <> 0 then
          begin
            WriteLn('  error reading structures');
            SetLength(FileList, 0);
            Exit
          end;

          if c <> #000 then
            Name := Name + c
        until c = #000
      end
    end
  until FileList[Idx].Offset = FileSize(PAKFile);

  for i := 0 to High(FileList) - 1 do
  begin
    with FileList[i] do
    begin
      Size := FileList[i + 1].Offset - Offset;

      WriteLn(Format('  %s @ %d up by %d bytes', [Name, Offset, Size]));

      {$I-}
      Seek(PAKFile, Offset);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    seek error');
        Continue
      end;

      if not OpenOutFile(OutFile, Name) then
      begin
        WriteLn('    error opening output file');
        Continue
      end;

      {$I-}
      for j := 1 to Size div SizeOf(Buffer) do
      begin
        BlockRead(PAKFile, Buffer, SizeOf(Buffer));
        BlockWrite(OutFile, Buffer, SizeOf(Buffer))
      end;
      BlockRead(PAKFile, Buffer, Size mod SizeOf(Buffer));
      BlockWrite(OutFile, Buffer, Size mod SizeOf(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    error copying data');
        CloseFile(OutFile);
      end;

      CloseFile(OutFile)
    end
  end;

  SetLength(FileList, 0)
end;

procedure ScanFiles(FileParam: AnsiString);
var
  Dir,
  InName   : AnsiString;
  InFile   : file;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      InName := Dir + SearchRec.Name;

      WriteLn(InName);

      if OpenInFile(InFile, InName) then
        ExtractFiles(InFile)
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  SysUtils.FindClose(SearchRec)
end;

begin
  if ParamCount = 1 then
    ScanFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.

